// This file is part of Awali.
// Copyright 2016-2020 Sylvain Lombardy, Victor Marsault, Jacques Sakarovitch
//
// Awali is a free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <http://www.gnu.org/licenses/>.

#ifndef CORA_EDIT_HH
#define CORA_EDIT_HH

#include <ncurses.h>
#include <unordered_map>
#include <cstring>
#include <stack>
#include <fstream>
#include <thread> // For sleep
#include <chrono> // 
#include <awali/dyn.hh>

#include <cora/bless.hh>
#include <cora/area.hh>
#include <cora/token.hh>

namespace awali { 
namespace cora {
namespace details { 


  // Shortcut hints
  typedef struct tip {
    std::string key;
    std::string hint;
  } tip_t;

  typedef std::vector<tip_t> tips_t;

  typedef enum input_kind { 
    NORMAL=0,              // char between 32 and 126
    UP, LEFT, DOWN, RIGHT, // arrow keys
    BACKSPACE, DELETE,     // suppr previous/next char
    BACK, FORWARD,         // ESCAPE/ENTER
    OTHER                  //
  } input_kind_t;


  typedef enum answer_kind {
    ABORT = 0,  // The user aborted
    INPUT = 1,  // The user provided an answer
    DEFAULT = 2 // Default behaviour should be 
  } answer_kind_t;


  // editstate_t is the different states in which the program might be
  // waiting for characters.
  typedef enum editstate {
    MAIN_STATE = 0,
    MAIN_TRANS = 1, 
    ASK = 2,
    TRANSEDIT = 3,
  } editstate_t;
  

  typedef struct tr_pack {
    int id; // id of the transition if it exists, -1 otherwise
    std::string src;
    std::string dst;
    std::string weight;
    std::vector<std::string> labels;
    const static std::string DEFAULT;
    
    tr_pack(dyn::automaton_t aut) :
      id(-1),
      src(DEFAULT),
      dst(DEFAULT),
      weight("")
    {
      if (aut->is_transducer())
        labels = std::vector<std::string>(
          ((dyn::transducer_t) aut).num_tapes(),DEFAULT);
      else
        labels = {DEFAULT};
      dyn::context_t ctx = aut->get_context();
      weight = ctx->weight_to_string(ctx->weight_one());
    }
    
    tr_pack(dyn::automaton_t aut, dyn::transition_t t, const std::string& eps_repr)
      :tr_pack(aut)
    {
      id = t;
      src = aut->get_state_name(aut->src_of(t));
      dst = aut->get_state_name(aut->dst_of(t));
      weight = aut->get_context()->weight_to_string(aut->weight_of(t));
      if (aut->is_transducer()) {
        labels.clear();
        for(const std::string& l : ((dyn::transducer_t) aut).label_of(t)) {
          labels.push_back(l==""?eps_repr:l);
        }
      }
      else {
        labels[0] =  
          aut->is_eps_transition(t) 
             ? eps_repr
             : aut->get_context()->label_to_string(aut->label_of(t));
      }
    }
  } tr_pack_t;
  
  class editter_t;

  class error_message_ostream_t : public std::ostream { 
    class custombuffer_t : public std::streambuf {
      std::stringstream ss;
      editter_t& e;
      public: custombuffer_t(editter_t& _e) : e(_e) {}
      protected: int overflow(int ch) override;
    };
  public:
    error_message_ostream_t (editter_t&e) : std::ostream(new custombuffer_t(e)) {}
    ~error_message_ostream_t() {delete(rdbuf());}
  };

  class editter_t {
    dyn::automaton_t aut;
    dyn::transducer_t tdc;
    dyn::context_t ctx;
    std::vector<std::vector<std::string>> alphabets;
    bool is_transducer;
    bool is_int_automaton;
    bool has_epsilon;
    bool is_over_B;
    std::string weightset_name;
    error_message_ostream_t ss;
    
    bool is_saved;
    void notify_modified();

    std::string filename = "example.js";
    std::vector<tips_t> tips; //indexed by editstate_t.

    std::vector<dyn::state_t> states;
    std::vector<dyn::transition_t> transitions;

    std::chrono::high_resolution_clock::time_point error_message_time;
    double error_message_timeout = 1500; //in ms
    
    dyn::state_t current_state();
    std::string current_state_name();
    dyn::transition_t current_transition();

    bool continue_edit = true;
    bool require_refresh = true;

    std::vector<tokens_t> transition_data;
    void clear_transition_data();
    tokens_t& get_transition_data (unsigned i);

    std::vector<tokens_t> state_data;
    void clear_state_data();
    tokens_t& get_state_data (unsigned i);
    
    std::string eps_repr = "_"; //Representation of epsilon
    std::string tdc_lbl_sta = "[";
    std::string tdc_lbl_sep = ",";
    std::string tdc_lbl_end = "]";
    std::string weight_sta = "<";
    std::string weight_end = ">";
    token_t ini_arr_format;
    token_t fin_arr_format ;
    token_t weight_format =   "<%s>";
    tokens_t label_format;
    token_t tr_sta; 
    token_t tr_end;
    tokens_t tr_format; 

  
/* w = width of terminal, 
   h = height of terminal
                        1                 1
                       >-<               >-<
         <-------------> <---------------> <-------------------------------->
            ~ w/4             ~ w/4                       ~w/2
        +---------------+-----------------+----------------------------------+
    ^   |               |                 |                                  |
    |   |               |                 |                                  |
    |   |               |                 |                                  |
    |   |               |                 |                                  |
    |   |               |                 |                                  |
h-4 |   |               |                 |                                  |
    |   |  infos_area   |   state_area    |            trans_area            |
    |   |               |                 |                                  |
    |   |               |                 |                                  |
    |   |               |                 |                                  |
    |   |               |                 |                                  |
    |   |               |                 |                                  |
  v v   |               |                 |                                  |
1 |     +---------------+-----------------+--------+-------------------------+
  ^     |          comment_area (1 line)           |                         |
        +------------------------------------------|  message_area (2 lines) |
  v     |           query_area (1 line)            |                         |
0 |     +------------------------------------------+-------------------------+
  ^     |                             tips_area                              |
        |                             (2 lines)                              |
        +--------------------------------------------------------------------+

         <----------------------------------------> <----------------------->
                            ~ 2 w/3                              ~w/3
                                                    >-<
                                                     1
*/
    area_t screen, 
           infos_area,
           query_area,
           comment_area,
           message_area,
           tips_area;
   
    list_area_t state_area, trans_area;


    input_kind_t wait_for_char(char& c); 
    answer_kind_t wait_for_string(std::string& s);
    
    class string_validator_t {
     public:
      virtual bool validate(const std::string& str) const = 0;
    };

    /* General function for asking information to the user.
     * A new string is asked to the user until given string passes the
     * given validator.
     * The field the user has to fill is initially containing the string
     * `result`.
     * If `wait_for_string` returns `DEFAULT`, `deflt` is copied in `result`.
     */
    bool ask_data (std::string& result, 
      const std::string& question, const std::string& default_value,
      const string_validator_t& validator, const std::string& error_msg );

    /* Ask a weight compatible with context in field ctx. */
    bool ask_weight (std::string& result, 
      const std::string& question, const std::string& def);
   
    /* Ask a label compatible with the ith-tape of the automaton. */
    bool ask_label (std::string& result, unsigned int i, 
      const std::string& question, const std::string& def);
    
    /* Ask a filename. No verification is done*/
    bool ask_filename (std::string& result, 
      const std::string& question, const std::string& def);

    /* Ask a state name. No verification is done*/
    bool ask_state_name (std::string& result,
      const std::string& question, const std::string& def);
    
    /* Ask the name of an existing state */
    bool ask_state (std::string& result,
      const std::string& question, const std::string& def);

    /* Ask confirmation for an action */
    bool ask_confirmation(bool& b, const std::string& question, bool def);

    token_t format_weight_of(dyn::weight_t tr, font_style_t fs = NO_STYLE);

    void initialize();
    void cleanup();

    void populate_alphabets();

    void refresh_infos(bool clear_data = false);
    void refresh_states(bool clear_data = false);
    void refresh_transitions(bool clear_data = false);
    void refresh_tips();
    void refresh_message();

    tokens_t state_display_name(dyn::state_t s, bool b = false);

    void compute_area_geometry();

    void action_refresh();
    
    void action_trans_add();
    void action_trans_delete();
    void action_trans_edit();

    void action_state_transfrom();
    void action_state_add();
    void action_state_rename();
    void action_state_delete();
    void action_state_initial();
    void action_state_final();
    void action_save();
    void action_quit();
    void action_add_letter();
    void action_display();

    void action_main_state();
    void action_main_trans();

//     void erase_area(area_t area, bool refresh_scr = false);
//     void print_tokens(area_t area,  int yoffset, int xoffset, std::vector<std::string> tokens, std::string sep = " "); 
    
    public:
    void print_message(const std::string&);
    void print_message(const tokens_t&);
    void print_error_message(const char* format, ...);
    private:
    std::stack<editstate_t> editstate_stack;
    editstate_t editstate();
    void enter_state(editstate_t);
    void leave_state();


    void apply(const tr_pack_t& tr);
    bool edit_tr_pack(tr_pack_t& tr, int off = 0) ; 

   public:
    editter_t(dyn::automaton_t a, std::string f);
    void run();


  }; //end of class editter_t


}}} // end of namespace awali::cora::details, awali::cora and awali


namespace awali {
namespace cora {

  // Main edit function, filename can be provided to save back on input file
  // If filename is "", it is understood that no filename are provided.
  void edit(dyn::automaton_t a, std::string filename = "");

}} //end of namespaces awali::cora and awali

#endif // CORA_EDIT_HH
